/* aes256.cpp -- Advanced Encryption Standard 256-bit encrypt/decrypt.
 *
 * STRICT_COMPILE -- the presence of these lines causes the makefile to turn on
 *		     most compiler warning flags.
 *
 * This file is from derived fromlibgcrypt-1.4.4/rijndael.c:
 *
 * Rijndael (AES) for GnuPG
 * Copyright (C) 2000, 2001, 2002, 2003, 2007,
 *               2008 Free Software Foundation, Inc.
 *
 * This file is part of Libgcrypt.
 *
 * Libgcrypt is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * Libgcrypt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, see <http://www.gnu.org/licenses/>.
 *******************************************************************
 * The code here is based on the optimized implementation taken from
 * http://www.esat.kuleuven.ac.be/~rijmen/rijndael/ on Oct 2, 2000,
 * which carries this notice:
 *------------------------------------------
 * rijndael-alg-fst.c   v2.3   April '2000
 *
 * Optimised ANSI C code
 *
 * authors: v1.0: Antoon Bosselaers
 *          v2.0: Vincent Rijmen
 *          v2.3: Paulo Barreto
 *
 * This code [the original, not this file] is placed in the public domain.
 *------------------------------------------
 *
 * The SP800-38a document is available at:
 *   http://csrc.nist.gov/publications/nistpubs/800-38a/sp800-38a.pdf
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "aes256.h"

typedef unsigned int u32;

// Hard-code constants for 256-bit encryption.
#define ROUNDS		14
#define KC		(256/32)
#define BLOCKSIZE       (128/8)

#undef ALIGNED
#define ALIGNED __attribute__((aligned(8)))

static uchar encryptKey[ROUNDS+1][4][4] ALIGNED;
static uchar decryptKey[ROUNDS+1][4][4] ALIGNED;

// All the numbers.
#include "rijndael-tables.h"

// aes256_setkey -- set the key.
// The parameter is a pointer to a 256-bit (32-byte) key.
void
aes256_setkey(const uchar * const key)
{
	// Set up the encryption key.
#define W encryptKey
	uchar k[KC][4] ALIGNED;
	for (int i = 0; i < 32 /* keylen */; i++) {
		k[i >> 2][i & 3] = key[i];
	}

	uchar tk[KC][4] ALIGNED;
	for (int i = KC-1; i >= 0; i--) {
		*(u32 *)tk[i] = *(u32 *)k[i];
	}
	int r = 0;
	int t = 0;

	// Copy values into round key array.
	for (int i = 0; i < KC && r < ROUNDS + 1; ) {
		for (; i < KC && t < 4; i++, t++) {
			*(u32 *)W[r][t] = *(u32 *)tk[i];
		}
		if (t == 4) {
			r++;
			t = 0;
		}
	}

	int rconpointer = 0;
	while (r < ROUNDS + 1) {

		// While not enough round key material calculated calculate
		// new values.
		tk[0][0] ^= S[tk[KC-1][1]];
		tk[0][1] ^= S[tk[KC-1][2]];
		tk[0][2] ^= S[tk[KC-1][3]];
		tk[0][3] ^= S[tk[KC-1][0]];
		tk[0][0] ^= rcon[rconpointer++];

		// We know that KC == 8, so:
		for (int i = 1; i < KC/2; i++) {
			*(u32 *)tk[i] ^= *(u32 *)tk[i-1];
		}
		tk[KC/2][0] ^= S[tk[KC/2 - 1][0]];
		tk[KC/2][1] ^= S[tk[KC/2 - 1][1]];
		tk[KC/2][2] ^= S[tk[KC/2 - 1][2]];
		tk[KC/2][3] ^= S[tk[KC/2 - 1][3]];
		for (int i = KC/2 + 1; i < KC; i++) {
			*(u32 *)tk[i] ^= *(u32 *)tk[i-1];
		}

		// Copy values into round key array.
		for (int i = 0; i < KC && r < ROUNDS + 1; ) {
			for (; i < KC && t < 4; i++, t++) {
				*(u32 *)W[r][t] = *(u32 *)tk[i];
			}
			if (t == 4) {
				r++;
				t = 0;
			}
		}
	}
#undef W

	// Set up the decryption key.
	memcpy(decryptKey, encryptKey, sizeof decryptKey);
#define W decryptKey
	for (r = 1; r < ROUNDS; r++) {
		for (int i = 0; i<4; i++) {
			uchar *w = W[r][i];
			*(u32 *)w = *(u32 *)U1[w[0]] ^ *(u32 *)U2[w[1]] ^
				    *(u32 *)U3[w[2]] ^ *(u32 *)U4[w[3]];

		}
	}
#undef W
}

// do_encrypt -- encrypt one block.
// A and B must be aligned on 4-byte boundaries.
// A and B may be the same.
static void
do_encrypt(
	uchar * const b,
	const uchar * const a)
{
#define rk encryptKey

// Select one of four bytes from a value:
#define B0(value) ((value) >> 24)
#define B1(value) (((value) >> 16) & 0xff)
#define B2(value) (((value) >> 8) & 0xff)
#define B3(value) ((value) & 0xff)

	uint temp0 = *(u32 *)(a   ) ^ *(u32 *)rk[0][0];
	uint temp1 = *(u32 *)(a+ 4) ^ *(u32 *)rk[0][1];
	uint temp2 = *(u32 *)(a+ 8) ^ *(u32 *)rk[0][2];
	uint temp3 = *(u32 *)(a+12) ^ *(u32 *)rk[0][3];

	uint b0 = *(u32 *)T1[B0(temp0)] ^ *(u32 *)T2[B1(temp1)] ^
		  *(u32 *)T3[B2(temp2)] ^ *(u32 *)T4[B3(temp3)];
	uint b1 = *(u32 *)T1[B0(temp1)] ^ *(u32 *)T2[B1(temp2)] ^
		  *(u32 *)T3[B2(temp3)] ^ *(u32 *)T4[B3(temp0)];
	uint b2 = *(u32 *)T1[B0(temp2)] ^ *(u32 *)T2[B1(temp3)] ^
		  *(u32 *)T3[B2(temp0)] ^ *(u32 *)T4[B3(temp1)];
	uint b3 = *(u32 *)T1[B0(temp3)] ^ *(u32 *)T2[B1(temp0)] ^
		  *(u32 *)T3[B2(temp1)] ^ *(u32 *)T4[B3(temp2)];

	for (int r = 1; r < ROUNDS-1; r++) {
		temp0 = b0 ^ *(u32 *)rk[r][0];
		temp1 = b1 ^ *(u32 *)rk[r][1];
		temp2 = b2 ^ *(u32 *)rk[r][2];
		temp3 = b3 ^ *(u32 *)rk[r][3];

		b0 = *(u32 *)T1[B0(temp0)] ^ *(u32 *)T2[B1(temp1)] ^
		     *(u32 *)T3[B2(temp2)] ^ *(u32 *)T4[B3(temp3)];
		b1 = *(u32 *)T1[B0(temp1)] ^ *(u32 *)T2[B1(temp2)] ^
		     *(u32 *)T3[B2(temp3)] ^ *(u32 *)T4[B3(temp0)];
		b2 = *(u32 *)T1[B0(temp2)] ^ *(u32 *)T2[B1(temp3)] ^
		     *(u32 *)T3[B2(temp0)] ^ *(u32 *)T4[B3(temp1)];
		b3 = *(u32 *)T1[B0(temp3)] ^ *(u32 *)T2[B1(temp0)] ^
		     *(u32 *)T3[B2(temp1)] ^ *(u32 *)T4[B3(temp2)];
	}

	// Last round is special.
	temp0 = b0 ^ *(u32 *)rk[ROUNDS-1][0];
	temp1 = b1 ^ *(u32 *)rk[ROUNDS-1][1];
	temp2 = b2 ^ *(u32 *)rk[ROUNDS-1][2];
	temp3 = b3 ^ *(u32 *)rk[ROUNDS-1][3];

	*(u32 *)(b   ) = (T1[B0(temp0)][1] << 24 | T1[B1(temp1)][1] << 16 |
			  T1[B2(temp2)][1] <<  8 | T1[B3(temp3)][1]) ^
			 *(u32 *)rk[ROUNDS][0];
	*(u32 *)(b+ 4) = (T1[B0(temp1)][1] << 24 | T1[B1(temp2)][1] << 16 |
			  T1[B2(temp3)][1] <<  8 | T1[B3(temp0)][1]) ^
			 *(u32 *)rk[ROUNDS][1];
	*(u32 *)(b+ 8) = (T1[B0(temp2)][1] << 24 | T1[B1(temp3)][1] << 16 |
			  T1[B2(temp0)][1] <<  8 | T1[B3(temp1)][1]) ^
			 *(u32 *)rk[ROUNDS][2];
	*(u32 *)(b+12) = (T1[B0(temp3)][1] << 24 | T1[B1(temp0)][1] << 16 |
			  T1[B2(temp1)][1] <<  8 | T1[B3(temp2)][1]) ^
			 *(u32 *)rk[ROUNDS][3];

#undef rk
}

// aes256_encrypt -- encrypt a block of data.
// Parameters:
// iv -- pointer to 128-bit (16-byte) initialization vector.
// buf -- pointer to data to be encrypted (must be 4-byte aligned).
// len -- number of bytes to encrypt (must be a multiple of 16).
void
aes256_encrypt(
	const uchar * const iv,
	uchar *buf,
	const uint len)
{
	// CBC (cipher-block chaining) mode.

	// The first plaintext block is XORed with the initialization vector.
	// Subsequent blocks are XORed with the previous ciphertext block.
	const uchar * xorblock = iv;

	for (uint i = 0; i<len; i += BLOCKSIZE) {

		// XOR the plaintext.
		for (uint j = 0; j < BLOCKSIZE; j += sizeof(uint)) {
			*(uint *)&buf[j] ^= *(uint *)&xorblock[j];
		}

		// Encrypt the block, turning plaintext into ciphertext.
		do_encrypt(buf, buf);

		xorblock = buf;

		// Advance to the next block.
		buf += BLOCKSIZE;
	}
}

// do_decrypt -- decrypt one block.
// A and B must be aligned on a 4-byte boundary.
// A and B may be the same.
static void
do_decrypt(
	uchar * const b,
	const uchar * const a)
{
#define rk decryptKey

	uint temp0 = *(u32 *)(a   ) ^ *(u32 *)rk[ROUNDS][0];
	uint temp1 = *(u32 *)(a+ 4) ^ *(u32 *)rk[ROUNDS][1];
	uint temp2 = *(u32 *)(a+ 8) ^ *(u32 *)rk[ROUNDS][2];
	uint temp3 = *(u32 *)(a+12) ^ *(u32 *)rk[ROUNDS][3];

	uint b0 = *(u32 *)T5[B0(temp0)] ^ *(u32 *)T6[B1(temp3)] ^
		  *(u32 *)T7[B2(temp2)] ^ *(u32 *)T8[B3(temp1)];
	uint b1 = *(u32 *)T5[B0(temp1)] ^ *(u32 *)T6[B1(temp0)] ^
		  *(u32 *)T7[B2(temp3)] ^ *(u32 *)T8[B3(temp2)];
	uint b2 = *(u32 *)T5[B0(temp2)] ^ *(u32 *)T6[B1(temp1)] ^
		  *(u32 *)T7[B2(temp0)] ^ *(u32 *)T8[B3(temp3)];
	uint b3 = *(u32 *)T5[B0(temp3)] ^ *(u32 *)T6[B1(temp2)] ^
		  *(u32 *)T7[B2(temp1)] ^ *(u32 *)T8[B3(temp0)];

	for (int r = ROUNDS-1; r > 1; r--) {
		temp0 = b0 ^ *(u32 *)rk[r][0];
		temp1 = b1 ^ *(u32 *)rk[r][1];
		temp2 = b2 ^ *(u32 *)rk[r][2];
		temp3 = b3 ^ *(u32 *)rk[r][3];
		b0 = *(u32 *)T5[B0(temp0)] ^ *(u32 *)T6[B1(temp3)] ^
		     *(u32 *)T7[B2(temp2)] ^ *(u32 *)T8[B3(temp1)];
		b1 = *(u32 *)T5[B0(temp1)] ^ *(u32 *)T6[B1(temp0)] ^
		     *(u32 *)T7[B2(temp3)] ^ *(u32 *)T8[B3(temp2)];
		b2 = *(u32 *)T5[B0(temp2)] ^ *(u32 *)T6[B1(temp1)] ^
		     *(u32 *)T7[B2(temp0)] ^ *(u32 *)T8[B3(temp3)];
		b3 = *(u32 *)T5[B0(temp3)] ^ *(u32 *)T6[B1(temp2)] ^
		     *(u32 *)T7[B2(temp1)] ^ *(u32 *)T8[B3(temp0)];
	}

	// Last round is special.
	temp0 = b0 ^ *(u32 *)rk[1][0];
	temp1 = b1 ^ *(u32 *)rk[1][1];
	temp2 = b2 ^ *(u32 *)rk[1][2];
	temp3 = b3 ^ *(u32 *)rk[1][3];

	*(u32 *)(b   ) = (S5[B0(temp0)] << 24 | S5[B1(temp3)] << 16 |
			  S5[B2(temp2)] <<  8 | S5[B3(temp1)]) ^
			 *(u32 *)rk[0][0];
	*(u32 *)(b+ 4) = (S5[B0(temp1)] << 24 | S5[B1(temp0)] << 16 |
			  S5[B2(temp3)] <<  8 | S5[B3(temp2)]) ^
			 *(u32 *)rk[0][1];

	*(u32 *)(b+ 8) = (S5[B0(temp2)] << 24 | S5[B1(temp1)] << 16 |
			  S5[B2(temp0)] <<  8 | S5[B3(temp3)]) ^
			 *(u32 *)rk[0][2];

	*(u32 *)(b+12) = (S5[B0(temp3)] << 24 | S5[B1(temp2)] << 16 |
			  S5[B2(temp1)] <<  8 | S5[B3(temp0)]) ^
			 *(u32 *)rk[0][3];

#undef rk
}

// aes256_decrypt -- decrypt a block of data.
// Parameters:
// iv -- pointer to 128-bit (16-byte) initialization vector.
// buf -- pointer to data to be decrypted (must be 4-byte aligned).
// len -- number of bytes to decrypt (must be a multiple of 16).
void
aes256_decrypt(
	const uchar * const iv,
	uchar *buf,
	const uint len)
{
	// CBC (cipher-block chaining) mode.

	// The first plaintext block (the result of the first block decryption)
	// is XORed with the initialization vector.
	// Subsequent plaintext blocks are XORed with the previous ciphertext
	// block.
	// We have to maintain two XOR blocks: the one we're about to use on the
	// current block, and a copy of the current ciphertext to be used with
	// the next block.
	// xorblock_i flips back and forth between 0 and 1, indicating the next
	// XOR block to be used.
	uchar xorblock[2][BLOCKSIZE] ALIGNED;
	memcpy(xorblock[0], iv, sizeof xorblock);
	uint xorblock_i = 0;

	for (uint i = 0; i<len; i += BLOCKSIZE) {

		// Take a copy of the ciphertext for use with the block
		// after this one.
		memcpy(xorblock[xorblock_i ^ 1], buf, BLOCKSIZE);

		// Decrypt the block, turning ciphertext into plaintext.
		do_decrypt(buf, buf);

		// XOR the plaintext.
		for (uint j = 0; j < BLOCKSIZE; j += sizeof(uint)) {
			*(uint *)&buf[j] ^= *(uint *)&xorblock[xorblock_i][j];
		}

		xorblock_i ^= 1;
		buf += BLOCKSIZE;
	}
}

#if 0 // include self tests

// aes256_selftest -- run the self tests for AES 256.
extern "C" int
aes256_selftest()
{
	uchar scratch[16] ALIGNED;

	static uchar plaintext_256[16] ALIGNED = {
		0x06,0x9A,0x00,0x7F,0xC7,0x6A,0x45,0x9F,
		0x98,0xBA,0xF9,0x17,0xFE,0xDF,0x95,0x21
	};
	static uchar key_256[32] ALIGNED = {
		0x08,0x09,0x0A,0x0B,0x0D,0x0E,0x0F,0x10,
		0x12,0x13,0x14,0x15,0x17,0x18,0x19,0x1A,
		0x1C,0x1D,0x1E,0x1F,0x21,0x22,0x23,0x24,
		0x26,0x27,0x28,0x29,0x2B,0x2C,0x2D,0x2E
	};
	static const uchar ciphertext_256[16] ALIGNED = {
		0x08,0x0E,0x95,0x17,0xEB,0x16,0x77,0x71,
		0x9A,0xCF,0x72,0x80,0x86,0x04,0x0A,0xE3
	};

	aes256_setkey(key_256);
	do_encrypt(scratch, plaintext_256);
	if (memcmp(scratch, ciphertext_256, sizeof ciphertext_256)) {
		printf("AES-256 test encryption failed.\n");
		return -1;
	}
	do_decrypt(scratch, scratch);
	if (memcmp(scratch, plaintext_256, sizeof plaintext_256)) {
		printf("AES-256 test decryption failed.\n");
		return -1;
	}

	printf("AES-256 test passed.\n");
	return 0;
}

#define SECTOR_LEN 512
static const char plaintext[SECTOR_LEN+1] =
	"press box\n"
	"E-mail Confidential\n"
	"Who's afraid of Time Inc.'s legal disclaimer?\n"
	"By Jack Shafer\n"
	"Posted Tuesday, June 1, 2004, at 3:42 PM PT\n"
	"\n"
	"The other day, a Time Inc. journalist of my acquaintance sent me an\n"
	"e-mail from his corporate e-mail account. I read it quickly and was\n"
	"about to hit the delete icon when I spotted this extraordinary\n"
	"114-word \"disclaimer\" sloshing around at the bottom. It read:\n"
	"\n"
	"    This message is the property of Time Inc. or its affiliates. It\n"
	"may be legally privileged and/or confidential ";

// aes256_bigtest -- encrypt and decrypt a disk-sector-size buffer.
extern "C" int
aes256_bigtest()
{
	static uchar sector[SECTOR_LEN];
	static const uchar iv[AES256_IV_LEN] = {
		0, 1, 1, 2, 3, 5, 8, 13, 21, 34, 55, 89, 144, 0xfe, 0xff
	};
	memcpy(sector, plaintext, SECTOR_LEN);
	aes256_encrypt(iv, sector, SECTOR_LEN);
	if (memcmp(sector, plaintext, SECTOR_LEN) == 0) {
		printf("AES-256 encryption didn't do anything?\n");
		return -1;
	}
	aes256_decrypt(iv, sector, SECTOR_LEN);
	if (memcmp(sector, plaintext, SECTOR_LEN) != 0) {
		printf("AES-256 big test failed\n");
		return -1;
	}
	printf("AES-256 big test passed.\n");
	return 0;
}

#endif
